import matplotlib.pyplot as plt
import numpy as np
from scipy import misc, stats
from sklearn.model_selection import train_test_split
import time


from CNN.csvData import CSVData
from CNN.relativeData import RelativeData, nparray2str

DENSO = 'cnn_data/merged_euler_densodataOVectors.csv'
OBJECTS = 'cnn_data/merged_positions_imagecenter.csv'


IMG_DEPTH = 'cnn_data/depth/'
# IMG_DEPTH = 'cnn_data/depth_RGB/'
# IMG_DEPTH = 'cnn_data/depth_GRAY/'
IMG_COLOR = 'cnn_data/color/'


class CNNData:
    def __init__(self, pixels=50, aug90 = True, aug180 = True):
        time_s = time.time()
        self.dimensions = (pixels, pixels, 4)
        self.pixels = pixels

        self.aug90 = aug90
        self.aug180 = aug180

        self.data_count = None
        self.batch_size = 25
        self.test_split = 0.25

        self.limit = 100
        self.norm_min = -250
        self.norm_max = 250

        self.indexes = None
        self.index_train = None
        self.index_test = None
        self.index_train_original = None
        self.index_test_original = None

        self.relativeData_object = None
        self.relative_6dof = None

        self.png_array = None
        self.image_count = None

        self.load_csv_data()
        self.train_test_split()
        #       self.png_to_nparray(IMG_DEPTH)
        #        self.png_to_nparray(IMG_COLOR)
        self.png_to_nparray_x(IMG_DEPTH, IMG_COLOR)

        self.augment_data()
        self.normalization_of_data()
        get_info(self)
        print("\nDATA LOADED: %.4fs" % (time.time() - time_s))

    def load_csv_data(self):
        print("**LOAD DATA**")
        salad_csv = CSVData(OBJECTS, densodata=False, example_count=self.data_count)
        gripper_csv = CSVData(DENSO, example_count=self.data_count)
        self.relativeData_object = RelativeData(salad_csv.coordinates, salad_csv.magnification)
        self.relativeData_object.set_gripperdata(gripper_csv.coordinates, gripper_csv.rotation, limit=self.limit)
        self.indexes = self.relativeData_object.get_indexes_under_limit()
        self.relative_6dof = self.relativeData_object.get_relative_6dof()
        self.data_count = len(self.relative_6dof)
        get_info(self.relativeData_object)

    def train_test_split(self):
        self.index_train, self.index_test = train_test_split(self.indexes, test_size=self.test_split,
                                                             random_state=42)
        self.index_test_original = self.index_test
        self.index_train_original = self.index_train
        print("\n**TRAIN & TEST SPLIT**\nTrain: %d \t Test: %d" % (len(self.index_train), len(self.index_test)))

    def png_to_nparray(self, folder):
        self.png_array = np.empty((tuple([len(self.relative_6dof)]) + self.dimensions))
        for i in range(len(self.relative_6dof)):
            png = misc.imread(folder + str(i + 1) + '.png')
            png = misc.imresize(png, self.dimensions)
            self.png_array[i] = png

    def png_to_nparray_x(self, folder_d, folder_c):
        self.png_array = np.empty((tuple([len(self.relative_6dof)]) + (self.pixels, self.pixels, 4)))
        for i in range(len(self.relative_6dof)):
            png_d = misc.imread(folder_d + str(i + 1) + '.png')
            png_d = misc.imresize(png_d, (self.pixels, self.pixels))

            png_c = misc.imread(folder_c + str(i + 1) + '.png')
            png_c = misc.imresize(png_c, (self.pixels, self.pixels, 3))

            self.png_array[i] = np.dstack((png_c, png_d))

    def get_training_batches(self):
        np.random.shuffle(self.index_train)

        # batch generated by definind how many in each batch
        rest = divmod(len(self.index_train), self.batch_size)

        split = [self.index_train[i * self.batch_size:(i + 1) * self.batch_size] for i in range(rest[0])]
        split.append(self.index_train[self.batch_size * rest[0]:])

        list_of_image_batches = []
        list_of_label_batches = []

        for b in split:
            image_batch = np.empty((tuple([len(b)]) + self.dimensions))
            label_batch = np.empty((len(b), 9))
            for i in range(len(b)):
                image_batch[i] = self.png_array[b[i]]
                label_batch[i] = self.relative_6dof[b[i]]

            list_of_image_batches.append(image_batch)
            list_of_label_batches.append(label_batch)

        return get_img_label_combo(image=list_of_image_batches, label=list_of_label_batches, split=split)

    def get_testing_batch(self):
        test_images = self.png_array[self.index_test]
        test_values = self.relative_6dof[self.index_test]

        return get_img_label_combo(image=test_images, label=test_values, split=self.index_test)

    def get_all_training(self):
        train_images = self.png_array[self.index_train]
        train_values = self.relative_6dof[self.index_train]

        return get_img_label_combo(image=train_images, label=train_values, split=self.index_train)

    def check_test_accuracy(self, output):
        difference = np.subtract(output, self.get_testing_batch()['label'])

        abs_angle = np.abs(difference[:, 3:])

        min_x = np.min(np.vstack(((np.subtract(360, abs_angle[:, 0])), abs_angle[:, 0])), axis=0)
        min_y = np.min(np.vstack(((np.subtract(360, abs_angle[:, 1])), abs_angle[:, 1])), axis=0)
        min_z = np.min(np.vstack(((np.subtract(180, abs_angle[:, 2])), abs_angle[:, 2])), axis=0)
        angle_diference = np.vstack((min_x, min_y, min_z)).T

        angle_distance = np.multiply(calculate_distance(angle_diference), 1)
        coord_distance = calculate_distance(difference[:, :3])
        accuracy = np.hstack((coord_distance, angle_distance))

        return difference, accuracy

    def augment_data(self):
        print("\n**AUGMENTING**\nInitial data size: %s [%s,%s]" % (
            len(self.relative_6dof), len(self.index_train), len(self.index_test)))
        aug_index_train = self.index_train
        aug_index_test = self.index_test
        aug_png_array = self.png_array
        aug_relative_6dof = self.relative_6dof

        if self.aug90:
            # Augment data by rotating each instance by 90, 180 and 270 degrees
            for i in range(1, 4):
                # augment data and images
                temp_aug_images, temp_aug_data = aug_90_rotate(self.png_array, self.relative_6dof, np.multiply(i, 90))
                aug_png_array = np.vstack((aug_png_array, temp_aug_images))
                aug_relative_6dof = np.vstack((aug_relative_6dof, temp_aug_data))
                # augment corresponding indexes
                aug_index_train = np.hstack((aug_index_train, np.add(self.index_train, self.data_count * i)))
                aug_index_test = np.hstack((aug_index_test, np.add(self.index_test, self.data_count * i)))

        if self.aug180:
            # AUGMENT BY ROTATING GRIPPER 180, image stays same
            aug_relative_6dof_flipped = augmentation_flip_grip(aug_relative_6dof)
            aug_png_array = np.vstack((aug_png_array, aug_png_array))
            aug_relative_6dof = np.vstack((aug_relative_6dof, aug_relative_6dof_flipped))

            aug_index_test = np.hstack((aug_index_test, np.add(aug_index_test, len(aug_relative_6dof_flipped))))
            aug_index_train = np.hstack((aug_index_train, np.add(aug_index_train, len(aug_relative_6dof_flipped))))

        self.index_train = aug_index_train
        self.index_test = aug_index_test
        self.png_array = aug_png_array
        self.relative_6dof = aug_relative_6dof

        print(
            "Augmented data size: %s [%s,%s]" % (len(self.relative_6dof), len(self.index_train), len(self.index_test)))
        # verification that correct data has been appended
        if False:
            print(self.index_train.shape)
            print(self.index_test.shape)
            print(self.png_array.shape)
            print(self.relative_6dof.shape)
            fig = plt.figure()
            for i in range(int(self.relative_6dof.shape[0] / self.data_count)):
                a = fig.add_subplot(2, 4, i + 1)
                index = i * (self.data_count)
                plt.imshow(self.png_array[index])
                print("Index: %s" % index)
                print("image smi unique value is: %s" % np.mean(self.png_array[index]))
                print(nparray2str([self.relative_6dof[index, :]]))
                a.set_title(nparray2str([self.relative_6dof[index, :]]))
            plt.show()

    def normalization_of_data(self):
        mean = np.mean(self.png_array)
        # self.png_array = np.subtract(self.png_array, mean)
        # self.png_array = np.divide(self.png_array, np.std(self.png_array))

        self.relative_6dof[:, :3] = norm(self.relative_6dof[:, :3], self.norm_min, self.norm_max)

        self.relative_6dof = deg2cos_sin(self.relative_6dof)
        self.relative_6dof[:, 3:] = norm(self.relative_6dof[:, 3:], -1, 1)

        # self.relative_6dof[:,3:] = norm(self.relative_6dof[:,3:], -180, 180)


def deg2cos_sin(six_dof):
    """
    :param six_dof: [X,Y,Z,Rx,Ry,Rz]
    :returns:   [X, Y, Z, cos(Rx), sin(Rx)...sin(Rz)]
    """
    rotation = six_dof[:, 3:]
    rotation = np.deg2rad(rotation)
    cos_sin = np.vstack((np.cos(rotation), np.sin(rotation))).reshape((-1, 6), order='F')
    return np.hstack((six_dof[:, :3], cos_sin))


def cos_sin2deg(nine_dof):
    """
    :param nine_dof: [X, Y, Z, cos(Rx), sin(Rx)...sin(Rz)]
    :return: [X,Y,Z,Rx,Ry,Rz]
    """
    #    cos = nine_dof[3::2]
    #    sin = nine_dof[4::2]
    nine_dof = np.reshape(nine_dof, [-1, 9])
    cos = nine_dof[:, [3, 5, 7]]
    sin = nine_dof[:, [4, 6, 8]]

    return np.hstack((nine_dof[:, :3], np.rad2deg(np.arctan2(sin, cos))))


def norm(x, min, max):
    return np.divide(np.subtract(x, min), np.subtract(max, min))


def denorm(x, min, max):
    return np.add(np.multiply(x, np.subtract(max, min)), min)


def aug_90_rotate(images, data, angle):
    print("AUGMENTNG BY: %3s degree rotation" % angle)
    rot90s = int(angle / 90)
    copy_images = np.copy(images)
    copy_data = np.copy(data)

    # rotate image by angle
    for i in range(len(copy_images)):
        copy_images[i] = misc.imrotate(copy_images[i], angle)

    # rotate data by angle (rotated 90 n times, where n is angle/90
    for _ in range(rot90s):
        # rotate coordinates by 90 degrees
        copy_data[:, [1, 0]] = copy_data[:, [0, 1]]
        copy_data[:, 0] = np.negative(copy_data[:, 0])
        # rotate rotateion by 90 degrees
        copy_data[:, 5] = np.where(np.add(copy_data[:, 5], 90) >= 180, np.subtract(copy_data[:, 5], 270),
                                   np.add(copy_data[:, 5], 90))

    return copy_images, copy_data


def augmentation_flip_grip(data_6dof):
    print("AUGMENTING BY: GRIPPER FLIP")
    flip_grip = np.copy(data_6dof)
    # negate Rx value
    flip_grip[:, 3] = np.multiply(flip_grip[:, 3], -1)
    # rotate Rz value 180 degrees and set to interval [-180,180]
    flip_grip[:, 5] = np.add(flip_grip[:, 5], 180)
    flip_grip[:, 5] = np.where(flip_grip[:, 5] >= 180, np.subtract(flip_grip[:, 5], 360), flip_grip[:, 5])
    return flip_grip


def calculate_distance(vectors):
    distance = np.square(vectors)
    distance = np.sum(distance, axis=1)
    return np.sqrt(distance).reshape(distance.shape[0], 1)


def get_img_label_combo(image, label, split):
    return {'img': image, 'label': label, 'split': split}


def get_info(object):
    print('\n**VARIABLES IN: %s**' % object.__class__.__name__)
    for label, value in vars(object).items():
        if isinstance(value, np.ndarray):
            print("%25s: %s" % (label, value.shape))
        else:
            print("%25s: %s" % (label, value))


def regression(goal, pred, text=""):
    goal_copy = np.copy(goal)
    pred_copy = np.copy(pred)

    data = []
    goal_copy[:, 3:] = np.add(goal_copy[:, 3:], 180)
    goal_copy[:, 5] = np.mod(goal_copy[:, 5], 180)
    pred_copy[:, 3:] = np.add(pred_copy[:, 3:], 180)
    pred_copy[:, 5] = np.mod(pred_copy[:, 5], 180)

    for i in range(goal_copy.shape[1]):
        denso_column = goal_copy[:, i]
        pred_column = pred_copy[:, i]

        column_regress = (stats.linregress(denso_column, pred_column)[2])

        rounded_column = round(column_regress, 6)

        data.append(float(rounded_column))

    print("%-25s" %text, str(data))
    return data


def save_prediction_to_file(predictions, filepath="testing.csv"):
    predictions = add_figure_and_preshape_data(predictions)
    predictions = add_labels_and_transform(predictions)
    np.savetxt(filepath, predictions, delimiter="\t", fmt="%s")
    print("Predictions saved to file: ", filepath)


def add_figure_and_preshape_data(data):
    if (data.shape[0] != 3):
        # Add figure 5.0
        data = np.hstack((data, np.full((data.shape[0], 1), 5)))

        # Add preshape 0.0
        data = np.hstack((data, np.full((data.shape[0], 1), 0)))

        # Add default fingerpressure 3x20.0
        data = np.hstack((data, np.full((data.shape[0], 3), 20)))
    rounded_data = np.round(data, decimals=6)
    return rounded_data

def add_labels_and_transform(data):
    label = np.array(["X", "Y", "Z", "RX", "RY", "RZ", "Figure", "Preshape", "Finger", "", ""])
    label = label.reshape(11, 1)
    labeled_data = np.hstack((label, np.transpose(data)))
    return labeled_data
