import numpy as np
import os
from scipy import stats
from sklearn.model_selection import train_test_split
from datetime import datetime


class IProcessData:
    # REGRESSION RESULTS FILE
    CSV_FILE = "data/regression_model.csv"

    #DENSO = "data/euler_densodataOVectors.csv"
    #FEATURES = "data/transformed_colordepth.csv"
    DENSO = "data/merged_data/euler_merged_densodataOVectors.csv"
    FEATURES = "data/merged_data/merged_transformed_features_colordepth.csv"


    # # DENSO
    # y_min = [360, -290, 113]
    # y_max = [827, 346, 143]
    # # OUTPUT
    # norm_out = [0, 1]

    def __init__(self, coordinates=False):
        self.single_counter = 0
        self.dataset_X = None
        self.dataset_y = None
        self.dataset_X_mean = None
        self.dataset_X_std = None
        self.dataset_y_mean = None
        self.dataset_y_std = None

        self.y_labels = None
        self.x_labels = None

        self.x_test_unaugmented = None
        self.y_test_unaugmented = None

        self.x_train = None
        self.y_train = None
        self.x_test = None
        self.y_test = None

        self.load_data()

        self.remove_figure_and_preshape()

        if coordinates:
            self.limit_to_coordinates()

        self.set_mean_std()

        print("-DATA LOADED")

    def load_data(self, densofile=DENSO, featurefile=FEATURES):
        """
        Loads the .csv data, as default the DENSO and TRANSFORMED data will be loaded
        Data loaded is located in same folder as the class, or else relative path is needed
        """
        self.x_labels, self.dataset_X = csv_to_nparray(featurefile)
        self.y_labels, self.dataset_y = csv_to_nparray(densofile)
        # self.dataset_y = self.dataset_y[:, :17]
        # self.y_labels = self.y_labels[:17]

    def remove_figure_and_preshape(self):
        if self.dataset_y.shape[1] == 17:
            self.dataset_y = np.delete(self.dataset_y, 12, axis=1)
            self.dataset_y = np.delete(self.dataset_y, 12, axis=1)

        if self.dataset_y.shape[1] == 11:
            self.dataset_y = np.delete(self.dataset_y, 6, axis=1)
            self.dataset_y = np.delete(self.dataset_y, 6, axis=1)

    def limit_to_coordinates(self):
        """
        Limits the data to only include positional coordinates X, Y, Z
        Allows for simple testing of positioning
        """
        self.dataset_X = self.dataset_X[:, :3]
        self.dataset_y = self.dataset_y[:, :3]

    def split_train_test(self, test_percentage=0.25):
        """
        Splits the loadded data into testing and training sets.
        test_size determines how large a persentage of the data is used for the test
        random_state uses the seed to allow for repetative consistent generation of splits
        """
        # TODO: SEED IS SET (random_state=0 // remove variable for randomrandomrandom)
        self.x_train, self.x_test, self.y_train, self.y_test = train_test_split(self.dataset_X, self.dataset_y,
                                                                                test_size=test_percentage,
                                                                                random_state=0)
        self.x_test_unaugmented = self.x_test
        self.y_test_unaugmented = self.y_test
        print("-DATA SPLIT: \t\ttest_size:\t", test_percentage)

    def augment_data(self, augmentation_count=15, augmentation_std=50):
        """
        uses augment_array method to increase the sizes of data
        """
        self.x_train, self.y_train = augment_array(self.x_train, self.y_train, augmentation_count, augmentation_std)
        self.x_test, self.y_test = augment_array(self.x_test, self.y_test, augmentation_count, augmentation_std)
        print("-DATA AUGMENTED; \ttraining:\t", len(self.x_train), "\ttest:\t", len(self.x_test))

    def limit_examples(self, count=None):
        """
        Limits the number of examples to allow for selective size testing
        starts at first example and expands
        could be improved to extract random set
        """
        if count is not None:
            self.dataset_X = self.dataset_X[:count, :]
            self.dataset_y = self.dataset_y[:count, :]

    def print_sizes(self):
        """
        Prints names and dimensions of all numpy arrays, useful for debugging
        """
        for i in self.__dict__:
            if type(self.__dict__.get(i)) is np.ndarray:
                print(i + ": shape:", self.__dict__.get(i).shape)

    # def output_to_norm(self):
    #     self.set_mean_std()
    #     for i in range(len(self.y_min)):
    #         self.y_train[:, i] = np.interp(self.y_train[:, i], [self.y_min[i], self.y_max[i]], self.norm_out)
    #         self.y_test[:, i] = np.interp(self.y_test[:, i], [self.y_min[i], self.y_max[i]], self.norm_out)
    #
    # def norm_to_output(self):
    #     for i in range(len(self.y_min)):
    #         self.y_train[:, i] = np.interp(self.y_train[:, i], self.norm_out, [self.y_min[i], self.y_max[i]])
    #         self.y_test[:, i] = np.interp(self.y_test[:, i], self.norm_out, [self.y_min[i], self.y_max[i]])
    #
    # def data_norm_to_output(self, data):
    #     data_T = data.T
    #     for i in range(len(self.y_min)):
    #         data_T[i] = np.interp(data_T[i], self.norm_out, [self.y_min[i], self.y_max[i]])
    #     return data_T.T

    def set_mean_std(self):
        self.dataset_X_mean = (np.mean(self.dataset_X, axis=0))
        self.dataset_X_std = (np.std(self.dataset_X, axis=0))
        self.dataset_y_mean = (np.mean(self.dataset_y, axis=0))
        self.dataset_y_std = (np.std(self.dataset_y, axis=0))


    def all_data_to_z_norm(self):

        self.x_train = convert_to_z_norm(self.x_train, self.dataset_X_mean, self.dataset_X_std)
        self.y_train = convert_to_z_norm(self.y_train, self.dataset_y_mean, self.dataset_y_std)

        self.x_test = convert_to_z_norm(self.x_test, self.dataset_X_mean, self.dataset_X_std)
        self.y_test = convert_to_z_norm(self.y_test, self.dataset_y_mean, self.dataset_y_std)
        self.x_test_unaugmented = convert_to_z_norm(self.x_test_unaugmented, self.dataset_X_mean, self.dataset_X_std)
        self.y_test_unaugmented = convert_to_z_norm(self.y_test_unaugmented, self.dataset_y_mean, self.dataset_y_std)

    def data_from_z_norm(self, data, input=False):
        # return data
        if input:
            return convert_from_z_norm(data, self.dataset_X_mean, self.dataset_X_std)
        else:
            return convert_from_z_norm(data, self.dataset_y_mean, self.dataset_y_std)

    def data_to_z_norm(self, data, input = False):
        # return data
        if input:
            return convert_to_z_norm(data, self.dataset_X_mean, self.dataset_X_std)
        else:
            return convert_to_z_norm(data, self.dataset_y_mean, self.dataset_y_std)



    def get_training_batches(self, num_of_batches=5):
        """
        Randomizes the training data and builds batches
        """
        idx = np.random.permutation(self.x_train.shape[0])
        temp_x = np.copy(self.x_train)
        temp_y = np.copy(self.y_train)
        temp_x = temp_x[idx, :]
        temp_y = temp_y[idx, :]

        if num_of_batches > len(self.x_train):
            print("ERROR: to many batches, defaulting to 5 batches")
            num_of_batches = 5
        return {'x': np.array_split(temp_x, num_of_batches), 'y': np.array_split(temp_y, num_of_batches)}

    def print_accuracy(self, acc, total_acc=True):
        if total_acc:
            print("Acc:\t", np2str(acc))
            print("AccSum:\t", round(np.mean(acc), 5))
        else:
            print("Acc\t", np2str(acc), 5)

    def print_max_min(self):
        i = 15
        print("np_X aka. Features")
        print(np.amin(self.dataset_X[:, :i], axis=0))
        print(np.amax(self.dataset_X[:, :i], axis=0))
        print("np_y aka. DENSO")
        print(np.amin(self.dataset_y[:, :i], axis=0))
        print(np.amax(self.dataset_y[:, :i], axis=0))
        if True:
            print("x_train")
            print(np.amin(self.x_train[:, :i], axis=0))
            print(np.amax(self.x_train[:, :i], axis=0))

            print("y_train")
            print(np.amin(self.y_train[:, :i], axis=0))
            print(np.amax(self.y_train[:, :i], axis=0))

            print("x_test")
            print(np.amin(self.x_test[:, :i], axis=0))
            print(np.amax(self.x_test[:, :i], axis=0))

            print("y_test")
            print(np.amin(self.y_test[:, :i], axis=0))
            print(np.amax(self.y_test[:, :i], axis=0))

    def get_single_test(self, index):
        """
        Gets a single example, usefull for predicting, as 1 set of features
        will be sent in to the NN and a prediction is expected in return
        """
        if index >= len(self.x_test):
            index = 0
            print("ERROR: index out of range")
        return {'x': self.x_test[index], 'y': self.y_test[index]}

    def get_single_data(self, index):
        return self.x_train[index], self.y_train[index]

    def save_prediction_to_file(self, predictions, filepath="testing.csv"):
        predictions = add_figure_and_preshape_data(predictions)
        predictions = add_labels_and_transform(predictions, self.y_labels)
        np.savetxt(filepath, predictions, delimiter="\t", fmt="%s")

    def save_regression_results(self, run_name, control, prediction):
        print("SAVING: Regression results to " + self.CSV_FILE)

        time = datetime.now().strftime('_%y%m%d_%H%M')
        regression_string = run_name + time
        data = []
        for i in range(control.shape[1]):
            denso_column = control[:, i]
            pred_column = prediction[:, i]

            column_regress = (stats.linregress(denso_column, pred_column)[2])

            rounded_column = round(column_regress, 6)

            regression_string += "\t" + str(rounded_column)
            data.append(float(rounded_column))

        with open(self.CSV_FILE, "a") as f:
            f.write(regression_string + "\n")

        print(data)


def convert_to_z_norm(data, mean, std):
    return np.divide(np.subtract(data, mean), std)


def convert_from_z_norm(data, mean, std):
    return np.add(np.multiply(data, std), mean)


def add_figure_and_preshape_data(data):
    if(data.shape[0]!=3):
        # Add figure
        data = np.insert(data, 6, 5, axis=0)
        # Add preshape
        data = np.insert(data, 7, 0, axis=0)
    rounded_data = np.round(data, decimals=6)

    return rounded_data


def add_labels_and_transform(data, label):

    if(data.shape[0]!=3):
        labeled_data = np.append(label.reshape(11, 1), data.reshape(11,data.shape[1]), axis=1)
    else:
        labeled_data = data
    return labeled_data


def augment_array(x, y, augmentation_count, augmentation_std):
    """
    Augments the data to allow for more data to be used in NN
    -augmentation_count for how many times to augment
    -augmentation_std to determine how broad an augmentation to use
    """
    aug_x = x
    aug_y = y
    x_shape = np.shape(x)
    y_shape = np.shape(y)
    # TODO: SEED IS SET
    np.random.seed(42)
    for n in range(augmentation_count):
        temp_random = augmentation_std * np.random.randn(x_shape[0], 3)

        temp_x = x + np.pad(temp_random, [(0, 0), (0, x_shape[1] - 3)], 'constant', constant_values=0)
        temp_y = y + np.pad(temp_random, [(0, 0), (0, y_shape[1] - 3)], 'constant', constant_values=0)

        aug_x = np.vstack((aug_x, temp_x))
        aug_y = np.vstack((aug_y, temp_y))

    return aug_x, aug_y


def csv_to_nparray(filename):
    """
    converts the .csv data to a numpy.nparray
    returns both Labels as L, and Values as V
    """
    file = open(filename, "r")
    data = [x.split("\t") for x in file]
    data_transformed = np.asarray(data).T

    labels = data_transformed[0, :]
    values = data_transformed[1:, :].astype(float)

    return labels, values


def np2str(data, precision=4):
    string = ""
    formater = "%8." + str(precision) + "f\t"
    for i in data:
        string += formater % i

    return string


def launch_tensorboard(LOG_DIRECTORY, chrome=False):
    temp = os.getcwd()
    os.chdir("D:\Anaconda\envs\python35")
    # START TENSORBOARD AND CHROME
    os.system("start python -m tensorflow.tensorboard --logdir=" + LOG_DIRECTORY)

    if chrome:
        os.system("start chrome http://192.168.1.201:6006/")
    os.chdir(temp)
